unit uTasks;
{$INCLUDE 'NamedFoldersLite.inc'}

interface

uses
  Windows, SysUtils, Classes,
  {$IFDEF FAR3} Plugin3, {$ELSE} PluginW, {$ENDIF} FarKeysW, FarColor,
  NativeXml,
  uLog, uFAR, uOptions, uMessages, uConsts, uXml, uFunctions,
  uEditDialog, uPanel, uInfo, uSubstitution;

type
  TNamedFolders = class
  private
  protected
    class var LastNamedFolderInfo: TNamedFolderInfo;
    class function ReadInfoFromPanel(PanelHandle: THandle; out Info: TNamedFolderInfo): boolean;
    class function DeleteInfoFromOptions(const Name: string): boolean;
    class function InfoExistsInOptions(const Name: string): boolean;
    class procedure InfoToDialog(Info: TNamedFolderInfo; Dialog: TEditNamedFolderDialog; out Directory: string);
    class procedure DialogToInfo(Dialog: TEditNamedFolderDialog; const Directory: string; Info: TNamedFolderInfo);
    class function GetSubstitutionEngines(out Engines: TSubstitutionEngines): boolean;
    class function ApplySubstitutionEngines(const FileName: string): string;
    class function SetCurrentDirToNamedFolder(Info: TNamedFolderInfo; PanelHandle: THandle; const InfoSuffix: string; AllowPluginCalls: boolean): boolean; overload;
  public
    class function ReadInfoFromOptions(const Name: string; out Info: TNamedFolderInfo; PartialMatchesAllowed: boolean = False): boolean;
    class function SaveInfoToOptions(const Info: TNamedFolderInfo): boolean;
    class function SetNamedFolder(Info: TNamedFolderInfo): boolean;
    class function SetNamedFolderToCurrentDir(const Name: string; PanelHandle: THandle): boolean; overload;
    class function SetNamedFolderToCurrentDir(const Name: string): boolean; overload;
    class function SetCurrentDirToNamedFolder(const Name: string; PanelHandle: THandle; PartialMatchesAllowed: boolean = False): boolean; overload;
    class function SetCurrentDirToNamedFolder(const Name: string; PartialMatchesAllowed: boolean = False): boolean; overload;
    class function ReadNamedFoldersList(const Directory: string; const Recursive: boolean; out List: TStringList; out DisplayNameStartIndex: integer): boolean;
    class function DeleteNamedFolder(const Name: string): boolean;
    class function ProcessCommandLine(const CommandLine: string): boolean;
    class function OpenPanel(const FocusedItem: string = ''): TNamedFoldersPanel;
    class function ExportNamedFolderToXml(Info: TNamedFolderInfo; Node: TXmlNode): boolean;
    class function ImportNamedFolderFromXml(Node: TXmlNode; out Info: TNamedFolderInfo): boolean;
    class function EditNamedFolder(const Name: string): boolean; overload;
    class function EditNamedFolder(Info: TNamedFolderInfo): boolean; overload;
    class function NewNamedFolder(const TargetDirectory: string; PanelHandle: THandle): boolean;
    class function MakeDirectory(const Path: string): boolean;
    class function HandleMacroCall(const Info: POpenMacroInfo): Pointer;
  end;

implementation

{ TNamedFolders }

class function TNamedFolders.ApplySubstitutionEngines(const FileName: string): string;
var
  Engines: TSubstitutionEngines;
begin
  Result := FileName;
  if GetSubstitutionEngines(Engines) then
    try
      Result := Engines.SubstituteVariables(Result);
    finally
      FreeAndNil(Engines);
    end;
end;

class function TNamedFolders.DeleteInfoFromOptions(const Name: string): boolean;
var
  Options: TOptions;
begin
  Options := TOptions.Create;
  try
    Result := Options.DeleteNamedFolderInfo(Name);
  finally
    FreeAndNil(Options);
  end;
end;

class function TNamedFolders.DeleteNamedFolder(const Name: string): boolean;
var
  Info: TNamedFolderInfo;
  OK: boolean;
  List: TStringList;
  FullName: string;
  i, DisplayNameStartIndex: integer;
begin
  Result := False;
  if not ReadInfoFromOptions(Name, Info, False) then
    FarMessage([MTitle, MFolderNotDefined, MOk], FMSG_WARNING, 1)
  else
    try
      OK := True;
      if Info.IsDirectory then begin
        FullName := TNamedFolderInfo.NormalizeName(Name);
        if ReadNamedFoldersList(FullName, True, List, DisplayNameStartIndex) then
          try
            for i := Pred(List.Count) downto 0 do
              if not DeleteInfoFromOptions(List[i]) then begin
                OK := False;
                Break;
              end;
          finally
            FreeAndNil(List);
          end;
      end;
      if OK then
        if DeleteInfoFromOptions(Name) then
          Result := True
        else
          FarMessage([MTitle, MDeletionFailed, MOk], FMSG_WARNING, 1);
    finally
      FreeAndNil(Info);
    end;
end;

class procedure TNamedFolders.DialogToInfo(Dialog: TEditNamedFolderDialog; const Directory: string; Info: TNamedFolderInfo);
begin
  if Directory = '' then
    Info.Name := Dialog.Name
  else
    Info.Name := Directory + '/' + Dialog.Name;
  Info.Directory := Dialog.Directory;
  Info.PluginID := Dialog.PluginID;
  Info.Param := Dialog.Param;
  Info.HostFile := Dialog.HostFile;
  Info.Description := Dialog.Description;
end;

class function TNamedFolders.EditNamedFolder(const Name: string): boolean;
var
  Info: TNamedFolderInfo;
begin
  Result := False;
  if ReadInfoFromOptions(Name, Info) then
    try
      Result := EditNamedFolder(Info);
    finally
      FreeAndNil(Info);
    end
end;

class function TNamedFolders.EditNamedFolder(Info: TNamedFolderInfo): boolean;
var
  OK: boolean;
  Dialog: TEditNamedFolderDialog;
  OriginalName, NewName, Directory: string;
begin
  Result := False;
  OriginalName := Info.Name;
  Dialog := TEditNamedFolderDialog.Create;
  try
    InfoToDialog(Info, Dialog, Directory);
    repeat
      OK := Dialog.Execute;
      if OK then
        if Dialog.Name = '' then begin
          FarMessage([MTitle, MNameMustNotBeEmpty, MOk], FMSG_WARNING, 1);
        end
        else begin
          DialogToInfo(Dialog, Directory, Info);
          NewName := Info.Name;
          SaveInfoToOptions(Info);
          if (OriginalName <> '') and (OriginalName <> NewName) then
            DeleteNamedFolder(OriginalName);
          Result := True;
        end;
    until Result or (not OK);
  finally
    FreeAndNil(Dialog);
    end;
end;

class function TNamedFolders.HandleMacroCall(const Info: POpenMacroInfo): Pointer;
const
  RESULT_FAILURE = Pointer(0);
  RESULT_SUCCESS = Pointer(1);

{$IFDEF FAR3_UP}

  procedure SyntaxError(const Command: string; Syntax: TMessages);
  var
    Msg: string;
  begin
    Msg := Format(GetMsgPas(MInvalidPluginCall), [Command]);
    FarMessage([MTitle, MExtraString1, Syntax, MOk], [PFarChar(Msg)], FMSG_WARNING, 1);
  end;

var
  Params: TVariantArray;
  ParamCount: integer;
  Command: string;

  procedure CmdOpen;
  var
    Panel: TNamedFoldersPanel;
  begin
    if ParamCount >= 2 then
      Panel := Self.OpenPanel(Params[1])
    else
      Panel := Self.OpenPanel;
    if Panel <> nil then
      Result := TFarUtils.DelphiToFarMacroCall([Panel]);
  end;

  procedure CmdGo;
  begin
    if ParamCount >= 2 then begin
      if SetCurrentDirToNamedFolder(Params[1], False) then
        Result := RESULT_SUCCESS;
    end
    else begin
      SyntaxError(Command, MInvalidPluginCallGoto);
    end;
  end;

  procedure CmdExists;
  begin
    if ParamCount >= 2 then begin
      if InfoExistsInOptions(Params[1]) then
        Result := RESULT_SUCCESS;
    end
    else begin
      SyntaxError(Command, MInvalidPluginCallExists);
    end;
  end;

  procedure CmdSetCurrent;
  begin
    if ParamCount >= 2 then begin
      if SetNamedFolderToCurrentDir(Params[1]) then
        Result := RESULT_SUCCESS;
    end
    else begin
      SyntaxError(Command, MInvalidPluginCallSetCurrent);
    end;
  end;

  procedure CmdSet;
  var
    FolderInfo: TNamedFolderInfo;
  begin
    if ParamCount >= 3 then begin
      FolderInfo := TNamedFolderInfo.Create;
      try
        FolderInfo.Name := Params[1];
        FolderInfo.Directory := Params[2];
        if ParamCount >= 4 then
          FolderInfo.Description := Params[3];
        if ParamCount >= 5 then
          FolderInfo.PluginID := StringToGUID(Params[4]);
        if ParamCount >= 6 then
          FolderInfo.Param := Params[5];
        if ParamCount >= 7 then
          FolderInfo.HostFile := Params[6];
        if SetNamedFolder(FolderInfo) then
          Result := RESULT_SUCCESS;
      finally
        FreeAndNil(FolderInfo);
      end;
    end
    else begin
      SyntaxError(Command, MInvalidPluginCallSet);
    end
  end;

  procedure CmdGet;
  var
    FolderInfo: TNamedFolderInfo;
    PluginID: string;
  begin
    if ParamCount >= 2 then begin
      if ReadInfoFromOptions(Params[1], FolderInfo, False) then begin
        try
          if IsEqualGUID(FolderInfo.PluginID, GUID_NULL) then
            PluginID := ''
          else
            PluginID := GUIDToString(FolderInfo.PluginID);
          Result := TFarUtils.DelphiToFarMacroCall([FolderInfo.Directory, FolderInfo.Description, PluginID, FolderInfo.Param, FolderInfo.HostFile]);
        finally
          FreeAndNil(FolderInfo);
        end;
      end;
    end
    else begin
      SyntaxError(Command, MInvalidPluginCallGet);
    end;
  end;

  procedure CmdDelete;
  begin
    if ParamCount >= 2 then begin
      if DeleteNamedFolder(Params[1]) then
        Result := RESULT_SUCCESS;
    end
    else begin
      SyntaxError(Command, MInvalidPluginCallDelete);
    end;
  end;

{$ENDIF}
begin
  Result := RESULT_FAILURE;
  {$IFDEF FAR3_UP}
  Params := TFarUtils.OpenMacroInfoToDelphi(Info);
  ParamCount := Length(Params);
  //Log(' - Params:');
  //for i := 0 to Pred(Length(Params)) do begin
  //  s := Params[i];
  //  Log('   - param %d: %s', [i, s]);
  //end;
  //Log(' - Params done');
  if ParamCount = 0 then begin
    CmdOpen;
  end
  else begin
    Command := AnsiUpperCase(Params[0]);
    if Command = 'OPEN' then begin
      CmdOpen;
    end
    else if (Command = 'GO') or (Command = 'GOTO') then begin
      CmdGo;
    end
    else if (Command = 'EXISTS') then begin
      CmdExists;
    end
    else if (Command = 'SETCURRENT') then begin
      CmdSetCurrent;
    end
    else if (Command = 'SET') then begin
      CmdSet;
    end
    else if (Command = 'GET') then begin
      CmdGet;
    end
    else if (Command = 'DELETE') then begin
      CmdDelete;
    end
    else begin
      FarMessage([MTitle, MInvalidPluginCommand, MExtraString1, MOk], [PFarChar(Command)], FMSG_WARNING, 1);
    end;
  end;
  {$ENDIF}
end;

class function TNamedFolders.ExportNamedFolderToXml(Info: TNamedFolderInfo; Node: TXmlNode): boolean;
begin
  Node.NodeNew(XML_ITEMNODE_NAME).ValueAsUnicodeString := GetLastPathElement(Info.Name);
  if Info.Description <> '' then
    Node.NodeNew(XML_ITEMNODE_DESCRIPTION).ValueAsUnicodeString := Info.Description;
  Node.NodeNew(XML_ITEMNODE_DIRECTORY).ValueAsUnicodeString := Info.Directory;
  if not IsEqualGUID(Info.PluginID, GUID_NULL) then
    Node.NodeNew(XML_ITEMNODE_PLUGINID).ValueAsUnicodeString := GUIDToString(Info.PluginID);
  if Info.Param <> '' then
    Node.NodeNew(XML_ITEMNODE_PARAM).ValueAsUnicodeString := Info.Param;
  if Info.HostFile <> '' then
    Node.NodeNew(XML_ITEMNODE_HOSTFILE).ValueAsUnicodeString := Info.HostFile;
  Result := True;
end;

class function TNamedFolders.GetSubstitutionEngines(out Engines: TSubstitutionEngines): boolean;

  procedure AddEngineIfAllowed(Allowed: boolean; EngineClass: TSubstitutionEngineClass; var WasSet: boolean);
  begin
    if Allowed then begin
      Engines.AddEngine(EngineClass);
      WasSet := True;
    end;
  end;

var
  Options: TOptions;
begin
  Result := False;
  Options := TOptions.Create;
  try
    if Options.UseSubstitutionEngines then begin
      Engines := TSubstitutionEngines.Create;
      try
        AddEngineIfAllowed(Options.UseSubstitutionEngine_Environment, TEnvVarSubstitutionEngine, Result);
        AddEngineIfAllowed(Options.UseSubstitutionEngine_BatchEnvironment, TBatchEnvVarSubstitutionEngine, Result);
        AddEngineIfAllowed(Options.UseSubstitutionEngine_KnownFolders, TKnownFolderSubstitutionEngine, Result);
        AddEngineIfAllowed(Options.UseSubstitutionEngine_KnownFolders, TKnownFolderSubstitutionEngine2, Result);
        AddEngineIfAllowed(Options.UseSubstitutionEngine_Registry, TRegistrySubstitutionEngine, Result);
        Result := True;
      except
        FreeAndNil(Engines);
        Result := False;
      end;
    end;
  finally
    FreeAndNil(Options);
  end;
end;

class function TNamedFolders.ImportNamedFolderFromXml(Node: TXmlNode; out Info: TNamedFolderInfo): boolean;
var
  Name, Directory, PluginIDStr: string;
begin
  Result := False;
  if not XmlValueByPath(Node, XML_ITEMNODE_NAME, Name) then
    FarMessage([MTitle, MMissingNameElement, MOk], FMSG_WARNING, 0)
  else if Name = '' then
    FarMessage([MTitle, MEmptyNameElement, MOk], FMSG_WARNING, 0)
  else if not XmlValueByPath(Node, XML_ITEMNODE_DIRECTORY, Directory) then
    FarMessage([MTitle, MMissingDirectoryElement, MOk], FMSG_WARNING, 0)
  else begin
    Info := TNamedFolderInfo.Create;
    try
      Info.Name := Name;
      Info.Description := XmlValueByPath(Node, XML_ITEMNODE_DESCRIPTION);
      Info.Directory := Directory;
      PluginIDStr := XmlValueByPath(Node, XML_ITEMNODE_PLUGINID);
      if PluginIDStr <> '' then
        Info.PluginID := StringToGUID(PluginIDStr);
      Info.Param := XmlValueByPath(Node, XML_ITEMNODE_PARAM);
      Info.HostFile := XmlValueByPath(Node, XML_ITEMNODE_HOSTFILE);
      Result := True;
    except
      FreeAndNil(Info);
    end;
  end;
end;

class function TNamedFolders.InfoExistsInOptions(const Name: string): boolean;
var
  Options: TOptions;
begin
  Options := TOptions.Create;
  try
    Result := Options.NamedFolderInfoExists(Name);
  finally
    FreeAndNil(Options);
  end;
end;

class procedure TNamedFolders.InfoToDialog(Info: TNamedFolderInfo; Dialog: TEditNamedFolderDialog; out Directory: string);
var
  Name: string;
begin
  Directory := Info.Name;
  if ExtractLastPathElement(Directory, Name) then
    Dialog.Name := Name
  else
    Dialog.Name := '';
  Dialog.Directory := Info.Directory;
  Dialog.PluginID := Info.PluginID;
  Dialog.Param := Info.Param;
  Dialog.HostFile := Info.HostFile;
  Dialog.Description := Info.Description;
end;

class function TNamedFolders.MakeDirectory(const Path: string): boolean;
var
  FullName, Element, DirName: string;
  Info: TNamedFolderInfo;
begin
  //Log('TNamedFolders.MakeDirectory("%s")', [Path]);
  Result := False;
  FullName := GetAbsoluteDirectory('', Path);
  DirName := '';
  //Log('- absolute name: "%s"', [FullName]);
  // I need to create the whole directory tree
  while ExtractPathElement(FullName, Element) do begin
    if Element <> '' then begin
      if DirName = '' then
        DirName := Element
      else
        DirName := DirName + '/' + Element;
      //Log('- creating "%s"', [DirName]);
      if TNamedFolders.ReadInfoFromOptions(DirName, Info, False) then begin
        //Log('- already exists');
        try
          if not Info.IsDirectory then begin
            FarMessage([MTitle, MFailedToCreateDir, MExtraString1, MOk], [PFarChar(DirName)], FMSG_WARNING, 1);
            Exit;
          end;
        finally
          FreeAndNil(Info);
        end;
      end
      else begin
        //Log('- doesn''t exist, creating');
        Info := TNamedFolderInfo.Create;
        try
          Info.Name := DirName;
          Info.IsDirectory := True;
          if not TNamedFolders.SaveInfoToOptions(Info) then begin
            FarMessage([MTitle, MFailedToCreateDir, MExtraString1, MOk], [PFarChar(DirName)], FMSG_WARNING, 1);
            Exit;
          end;
        finally
          FreeAndNil(Info);
        end;
      end;
    end;
  end;
  Result := True;
  //Log('TNamedFolders.MakeDirectory("%s") -> %d', [Path, Integer(Result)]);
end;

class function TNamedFolders.NewNamedFolder(const TargetDirectory: string; PanelHandle: THandle): boolean;
var
  OldInfo, Info: TNamedFolderInfo;
  OK, Again: boolean;
  Dialog: TEditNamedFolderDialog;
  DeleteName, Dummy: string;
begin
  Result := False;
  if ReadInfoFromPanel(PanelHandle, Info) then
    try
      repeat
        Dialog := TEditNamedFolderDialog.Create;
        try
          InfoToDialog(Info, Dialog, Dummy);
          Again := False;
          OK := Dialog.Execute;
          if OK then
            if Dialog.Name = '' then begin
              FarMessage([MTitle, MNameMustNotBeEmpty, MOk], FMSG_WARNING, 1);
              Again := True;
            end
            else begin
              DialogToInfo(Dialog, TargetDirectory, Info);
              if ReadInfoFromOptions(Info.Name, OldInfo) then
                try
                  if OldInfo.IsDirectory then begin
                    FarMessage([MTitle, MFolderAlreadyExistsAndCantBeOverwritten, MOk], FMSG_WARNING, 1);
                    OK := False;
                  end
                  else
                  case FarMessage([MTitle, MFolderAlreadyExists, MExtraString1, MFolderAlreadyExistsQuestion, MYes, MNo, MCancel], [PFarChar(OldInfo.Directory)], FMSG_WARNING, 3) of
                    0:   ; // do nothing
                    1:   Again := True;
                    else OK := False;
                  end;
                finally
                  FreeAndNil(OldInfo);
                end;
              if OK and (not Again) then begin
                SaveInfoToOptions(Info);
                if DeleteName <> '' then
                  DeleteNamedFolder(DeleteName);
                Result := True;
              end;
            end;
        finally
          FreeAndNil(Dialog);
        end;
      until not Again;
    finally
      FreeAndNil(Info);
    end;
end;

class function TNamedFolders.OpenPanel(const FocusedItem: string): TNamedFoldersPanel;
var
  Directory: string;
begin
  Result := TNamedFoldersPanel.Create;
  if FocusedItem <> '' then begin
    if not Result.ChangeDirectory(FocusedItem) then begin
      Directory := GetPath(FocusedItem);
      if Result.ChangeDirectory(Directory) then begin
        //Result.SelectFile(FocusedItem); // Doesn't work here, needs to be done later
        Result.FileToBeSelected := FocusedItem;
      end;
    end;
  end;
end;

class function TNamedFolders.ProcessCommandLine(const CommandLine: string): boolean;
begin
  Result := False;
  if CommandLine <> '' then
    if CommandLine[1] = ':' then
      Result := SetNamedFolderToCurrentDir(Trim(Copy(CommandLine, 2, MaxInt)))
    else
      Result := SetCurrentDirToNamedFolder(Trim(CommandLine), True);
end;

class function TNamedFolders.ReadInfoFromOptions(const Name: string; out Info: TNamedFolderInfo; PartialMatchesAllowed: boolean): boolean;

  function StartsWith(const Haystack, Needle: string): boolean;
    begin
      Result := Copy(Haystack, 1, Length(Needle)) = Needle;
    end;

var
  Options: TOptions;
  TempInfo: TNamedFolderInfo;
  List, MenuList: TStringList;
  FinalName, MenuListFormat, TempName, ExactMatchOverDirs, NameWithoutDirs: string;
  i, ItemIndex, NameLength, MaxNameLength: integer;
  KeepItem, AllowFuzzySearch, AllowSearchOverDirectories, ExactMatchOverDirsFound: boolean;
begin
  Result := False;
  FinalName := TNamedFolderInfo.NormalizeName(Name);
  Options := TOptions.Create;
  try
    AllowFuzzySearch := Options.AllowFuzzySearch;
    AllowSearchOverDirectories := Options.AllowSearchOverDirectories;
    if PartialMatchesAllowed and (AllowFuzzySearch or AllowSearchOverDirectories) then begin
      Options.NamedFoldersList(List);
      try
        ExactMatchOverDirs := '';
        ExactMatchOverDirsFound := False;
        for i := Pred(List.Count) downto 0 do begin
          if List[i] = FinalName then begin
            Result := Options.LoadNamedFolderInfo(FinalName, Info);
            Exit;
          end
          else begin
            if AllowSearchOverDirectories then begin
              NameWithoutDirs := GetLastPathElement(List[i]);
              if NameWithoutDirs = FinalName then begin
                KeepItem := True;
                if ExactMatchOverDirsFound then begin
                  ExactMatchOverDirs := '';
                end
                else begin
                  ExactMatchOverDirsFound := True;
                  ExactMatchOverDirs := List[i];
                end;
              end
              else begin
                KeepItem := AllowFuzzySearch and StartsWith(NameWithoutDirs, FinalName);
              end;
            end
            else if AllowFuzzySearch and StartsWith(List[i], FinalName) then begin
              KeepItem := True;
            end
            else begin
              KeepItem := False;
            end;
            if not KeepItem then
              List.Delete(i);
          end;
        end;
        if ExactMatchOverDirs <> '' then begin
          Result := Options.LoadNamedFolderInfo(ExactMatchOverDirs, Info);
          Exit;
        end;
        if List.Count > 0 then begin
          if List.Count = 1 then begin
            Result := Options.LoadNamedFolderInfo(List[0], Info);
          end
          else begin
            MenuList := TStringList.Create;
            try
              MaxNameLength := 0;
              for i := 0 to Pred(List.Count) do begin
                NameLength := Length(List[i]);
                if NameLength > MaxNameLength then
                  MaxNameLength := NameLength;
              end;
              MenuListFormat := Format('%%%0:d.%0:ds   %%s', [MaxNameLength]);
              for i := 0 to Pred(List.Count) do begin
                TempName := List[i];
                if Options.LoadNamedFolderInfo(TempName, TempInfo) then begin
                  try
                    if not TempInfo.IsDirectory then
                      MenuList.Add(Format(MenuListFormat, [TempName, TempInfo.DisplayableDescription]));
                  finally
                    FreeAndNil(TempInfo);
                  end;
                end;
              end;
              if MenuList.Count > 0 then
                if MenuList.Count = 1 then
                  Result := Options.LoadNamedFolderInfo(List[0], Info)
                else
                  if TFarUtils.Menu(MENU_CHOOSE_NAME, GetMsg(MChooseName), 'ChooseNameMenu', MenuList, ItemIndex) then begin
                    Result := Options.LoadNamedFolderInfo(List[ItemIndex], Info);
                  end;
            finally
              FreeAndNil(MenuList);
            end;
          end;
        end;
      finally
        FreeAndNil(List);
      end;
    end
    else begin
      Result := Options.LoadNamedFolderInfo(FinalName, Info);
    end;
  finally
    FreeAndNil(Options);
  end;
end;

class function TNamedFolders.ReadInfoFromPanel(PanelHandle: THandle; out Info: TNamedFolderInfo): boolean;
var
  Panel: TFarPanelInfo;
begin
  Result := False;
  Panel := TFarPanelInfo.Create(PanelHandle, False);
  try
    Info := TNamedFolderInfo.Create;
    try
      Info.Name := '';
      Info.Directory := Panel.Directory;
      Info.PluginID := {$IFDEF FAR3_UP} Panel.DirectoryPlugin {$ELSE} NULL_GUID {$ENDIF} ;
      Info.Param := {$IFDEF FAR3_UP} Panel.DirectoryParam {$ELSE} '' {$ENDIF} ;
      Info.HostFile := {$IFDEF FAR3_UP} Panel.DirectoryFile {$ELSE} '' {$ENDIF} ;
      Result := True;
    except
      FreeAndNil(Info);
      Raise;
    end;
  finally
    FreeAndNil(Panel);
  end;
end;

class function TNamedFolders.ReadNamedFoldersList(const Directory: string; const Recursive: boolean; out List: TStringList; out DisplayNameStartIndex: integer): boolean;
var
  Options: TOptions;
  AbsDir, NamePrefix, FileName, FileNameEnd, DisplayFileName: string;
  NamePrefixLength: integer;
  i: Integer;
begin
  //Log('TNamedFolders.ReadNamedFoldersList("%s", %d)', [Directory, Integer(Recursive)]);
  Result := True;
  Options := TOptions.Create;
  try
    Options.NamedFoldersList(List);
    try
      AbsDir := GetAbsoluteDirectory('', Directory);
      if AbsDir <> '' then begin
        NamePrefix := AbsDir + '/';
        NamePrefixLength := Length(NamePrefix);
      end
      else begin
        NamePrefix := '';
        NamePrefixLength := 0;
      end;
      //Log('- Directory: "%s"', [NamePrefix]);
      DisplayNameStartIndex := Succ(NamePrefixLength);
      for i := Pred(List.Count) downto 0 do begin
        FileName := List[i];
        //Log('- processing "%s"', [FileName]);
        // File's first part must match the directory
        if (NamePrefixLength = 0) or (Copy(FileName, 1, NamePrefixLength) = NamePrefix) then begin
          //Log('- directory match 1');
          // At least something must be left
          FileNameEnd := Copy(FileName, Succ(NamePrefixLength), MaxInt);
          //Log('- testing DisplayFileName "%s"', [FileNameEnd]);
          if ExtractPathElement(FileNameEnd, DisplayFileName) then begin
            // But that something must not contain another path element, or Recursive must be active
            //Log('- displayname match: "%s"', [DisplayFileName]);
            if Recursive or (FileNameEnd = '') then begin
              // Got a Name which matches the requirements
              //Log('- directory match 2');
              Continue;
            end;
          end;
        end;
        // Apparently the name doesn't match the directory, delete it.
        //Log('- deleting "%s"', [FileName]);
        List.Delete(i);
      end;
    except
      FreeAndNil(List);
      Raise;
    end;
  finally
    FreeAndNil(Options);
  end;
  //Log('TNamedFolders.ReadNamedFoldersList("%s", %d) -> %d, %d items', [Directory, Integer(Recursive), Integer(Result), List.Count]);
end;

class function TNamedFolders.SaveInfoToOptions(const Info: TNamedFolderInfo): boolean;
var
  Options: TOptions;
begin
  Options := TOptions.Create;
  try
    Result := Options.SaveNamedFolderInfo(Info);
  finally
    FreeAndNil(Options);
  end;
end;

class function TNamedFolders.SetCurrentDirToNamedFolder(Info: TNamedFolderInfo; PanelHandle: THandle; const InfoSuffix: string; AllowPluginCalls: boolean): boolean;
var
  Directory, LastDirectory: string;
  Again, SearchForNearestDirActive: boolean;
begin
  Result := False;
  if Info.IsDirectory then
    FarMessage([MTitle, MFolderNotDefined, MOk], FMSG_WARNING, 1)
  else begin
    SearchForNearestDirActive := False;
    Directory := Info.Directory;
    if IsEqualGUID(Info.PluginID, NULL_GUID) then
      Directory := ApplySubstitutionEngines(Directory);
    Directory := Directory + InfoSuffix;
    // For some reason, FAR will refuse to set a panel directory to \\path,
    // even if it will work fine with CD \\path.
    if IsEqualGUID(Info.PluginID, NULL_GUID) and (Length(Directory) > 2) and (Directory[1] = '\') and (Directory[2] = '\') then begin
      Result := TFarUtils.ExecuteCommandLine('CD ' + AnsiQuotedStr(Directory, '"'));
    end
    else if AllowPluginCalls and IsEqualGUID(Info.PluginID, NULL_GUID) and TFarUtils.FindPlugin(Directory) then begin
      Result := TFarUtils.ExecuteCommandLine(Directory);
    end
    else repeat
      Again := False;
      Result := TFarUtils.SetCurrentDir(Directory, Info.Param, Info.HostFile, Info.PluginID, PanelHandle);
      if not Result then
        if IsEqualGUID(Info.PluginID, NULL_GUID) then
          if SearchForNearestDirActive or (FarMessage([MTitle, MFolderDoesNotExist, MExtraString1, MGoToNearestDirQuestion, MYes, MNo, MCancel], [PFarChar(Directory)], FMSG_WARNING, 3) = 0) then begin
            SearchForNearestDirActive := True;
            LastDirectory := Directory;
            Directory := ExcludeTrailingPathDelimiter(ExtractFilePath(Directory));
            Again := (Directory <> '') and (Directory <> LastDirectory);
            if not Again then
              FarMessage([MTitle, MFolderDoesNotExist, MOK], FMSG_WARNING, 1);
          end;
    until not Again;
  end;
end;

class function TNamedFolders.SetCurrentDirToNamedFolder(const Name: string; PanelHandle: THandle; PartialMatchesAllowed: boolean): boolean;
var
  Options: TOptions;
  AllowPathSuffix, AllowPluginCalls: boolean;
  Info: TNamedFolderInfo;
  InfoName, InfoSuffix: string;
  ix: integer;
  CurrentInfo: TNamedFolderInfo;
begin
  Options := TOptions.Create;
  try
    AllowPathSuffix := Options.AllowPathSuffix;
    AllowPluginCalls := Options.AllowPluginCalls;
  finally
    FreeAndNil(Options);
  end;
  Result := False;
  InfoName := Name;
  InfoSuffix := '';
  if AllowPathSuffix then begin
    ix := Pos('\', InfoName);
    if ix > 0 then begin
      InfoSuffix := Copy(InfoName, ix, MaxInt);
      SetLength(InfoName, Pred(ix));
    end;
  end;
  if not ReadInfoFromPanel(PANEL_ACTIVE, CurrentInfo) then
    CurrentInfo := nil;
  try
    if InfoName = '-' then
    begin
      if Assigned(LastNamedFolderInfo) then
        Result := SetCurrentDirToNamedFolder(LastNamedFolderInfo, PANEL_ACTIVE, InfoSuffix, True)
      else
        FarMessage([MTitle, MLastNamedFolderInfoDoesNotExist, MOK], FMSG_WARNING, 1);
    end
    else
    if ReadInfoFromOptions(InfoName, Info, PartialMatchesAllowed) then
      try
        Result := SetCurrentDirToNamedFolder(Info, PanelHandle, InfoSuffix, AllowPluginCalls);
      finally
        FreeAndNil(Info);
      end
    else
      FarMessage([MTitle, MFolderNotDefined, MOk], FMSG_WARNING, 1);
  finally
    if Result then
    begin
      FreeAndNil(LastNamedFolderInfo);
      LastNamedFolderInfo := CurrentInfo;
    end
    else
    begin
      FreeAndNil(CurrentInfo);
    end;
  end;
end;

class function TNamedFolders.SetCurrentDirToNamedFolder(const Name: string; PartialMatchesAllowed: boolean): boolean;
begin
  Result := SetCurrentDirToNamedFolder(Name, PANEL_ACTIVE, PartialMatchesAllowed);
end;

class function TNamedFolders.SetNamedFolderToCurrentDir(const Name: string; PanelHandle: THandle): boolean;
var
  CurrentInfo, OldInfo: TNamedFolderInfo;
begin
  //Log('TNamedFolders.SetNamedFolderToCurrentDir("%s")', [Name]);
  Result := False;
  if ReadInfoFromPanel(PanelHandle, CurrentInfo) then
    try
      CurrentInfo.Name := Name;
      if ReadInfoFromOptions(Name, OldInfo, False) then
        CurrentInfo.Description := OldInfo.Description;
      Result := SetNamedFolder(CurrentInfo);
    finally
      FreeAndNil(CurrentInfo);
    end;
  //Log('TNamedFolders.SetNamedFolderToCurrentDir("%s") -> %d', [Name, Integer(Result)]);
end;

class function TNamedFolders.SetNamedFolder(Info: TNamedFolderInfo): boolean;
var
  OldInfo: TNamedFolderInfo;
begin
  Result := False;
  if not MakeDirectory(GetPath(Info.Name)) then
    Exit;
  if ReadInfoFromOptions(Info.Name, OldInfo) then begin
    try
      if OldInfo.IsDirectory then begin
        FarMessage([MTitle, MFolderAlreadyExistsAndCantBeOverwritten, MOk], FMSG_WARNING, 1);
        Exit;
      end
      else if IsEqualGUID(OldInfo.PluginID, NULL_GUID) then begin
        if FarMessage([MTitle, MFolderAlreadyExists, MExtraString1, MFolderAlreadyExistsQuestion, MYes, MNo, MCancel], [PFarChar(OldInfo.Directory)], FMSG_WARNING, 3) <> 0 then
          Exit;
      end
      else begin
        if FarMessage([MTitle, MFolderAlreadyExists, MExtraString1, MExtraString2, MExtraString3, MExtraString4, MFolderAlreadyExistsQuestion, MYes, MNo, MCancel], [PFarChar(OldInfo.Directory), PFarChar(GUIDToString(OldInfo.PluginID)), PFarChar(OldInfo.Param), PFarChar(OldInfo.HostFile)], FMSG_WARNING, 3) <> 0 then
          Exit;
      end;
    finally
      FreeAndNil(OldInfo);
    end;
  end;
  //Log('- saving to options');
  SaveInfoToOptions(Info);
  Result := True;
end;

class function TNamedFolders.SetNamedFolderToCurrentDir(const Name: string): boolean;
begin
  Result := SetNamedFolderToCurrentDir(Name, PANEL_ACTIVE);
end;

end.
